__all__ = [
    "AsDataArray",
    "AsDataset",
    "AsDataTree",
    "asdataarray",
    "asdataset",
    "asdatatree",
]

# standard library
from collections.abc import Callable
from typing import Any, Protocol, TypeVar, overload

# dependencies
import xarray as xr
from typing_extensions import ParamSpec
from .convert import to_dataarray, to_dataset, to_datatree, to_specframe

# type hints
P = ParamSpec("P")
T = TypeVar("T", bound=xr.DataArray | xr.Dataset | xr.DataTree)


class HasType(Protocol[P, T]):
    type: Callable[..., T]

    def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None: ...


class Other(Protocol[P]):

    def __init__(self, *args: P.args, **kwargs: P.kwargs) -> None: ...


class AsDataArray:
    """Mixin class for Xarray DataArray specifications."""

    @overload
    @classmethod
    def new(cls: type[HasType[P, T]], *args: P.args, **kwargs: P.kwargs) -> T: ...
    @overload
    @classmethod
    def new(cls: type[Other[P]], *args: P.args, **kwargs: P.kwargs) -> xr.DataArray: ...
    @classmethod
    def new(cls: Any, *args: Any, **kwargs: Any) -> Any:
        """Convert the Xarray specifications to an Xarray DataArray."""
        return asdataarray(cls(*args, **kwargs))


class AsDataset:
    """Mixin class for Xarray Dataset specifications."""

    @overload
    @classmethod
    def new(cls: type[HasType[P, T]], *args: P.args, **kwargs: P.kwargs) -> T: ...
    @overload
    @classmethod
    def new(cls: type[Other[P]], *args: P.args, **kwargs: P.kwargs) -> xr.Dataset: ...
    @classmethod
    def new(cls: Any, *args: Any, **kwargs: Any) -> Any:
        """Convert the Xarray specifications to an Xarray Dataset."""
        return asdataset(cls(*args, **kwargs))


class AsDataTree:
    """Mixin class for Xarray DataTree specifications."""

    @overload
    @classmethod
    def new(cls: type[HasType[P, T]], *args: P.args, **kwargs: P.kwargs) -> T: ...
    @overload
    @classmethod
    def new(cls: type[Other[P]], *args: P.args, **kwargs: P.kwargs) -> xr.DataTree: ...
    @classmethod
    def new(cls: Any, *args: Any, **kwargs: Any) -> Any:
        """Convert the Xarray specifications to an Xarray DataTree."""
        return asdatatree(cls(*args, **kwargs))


@overload
def asdataarray(obj: HasType[P, T], /) -> T: ...  # type: ignore
@overload
def asdataarray(obj: Other[P], /) -> xr.DataArray: ...
def asdataarray(obj: Any, /) -> Any:
    """Convert given Xarray specifications to an Xarray DataArray."""
    return to_dataarray(to_specframe(obj))


@overload
def asdataset(obj: HasType[P, T], /) -> T: ...  # type: ignore
@overload
def asdataset(obj: Other[P], /) -> xr.Dataset: ...
def asdataset(obj: Any, /) -> Any:
    """Convert given Xarray specifications to an Xarray Dataset."""
    return to_dataset(to_specframe(obj))


@overload
def asdatatree(obj: HasType[P, T], /) -> T: ...  # type: ignore
@overload
def asdatatree(obj: Other[P], /) -> xr.DataTree: ...
def asdatatree(obj: Any, /) -> Any:
    """Convert given Xarray specifications to an Xarray DataTree."""
    return to_datatree(to_specframe(obj))



# dependencies
import xarrayspecs as xs


def test_namespace() -> None:
    # submodules
    xs.api
    xs.convert
    xs.spec
    # aliases
    xs.AsDataArray
    xs.AsDataset
    xs.AsDataTree
    xs.Dims
    xs.Dtype
    xs.Type
    xs.asdataarray
    xs.asdataset
    xs.asdatatree
    xs.attrs
    xs.dims
    xs.dtype
    xs.name
    xs.node
    xs.type
    xs.use



__all__ = [
    "to_attrs",
    "to_coords",
    "to_data",
    "to_dataarray",
    "to_dataset",
    "to_datatree",
    "to_dims",
    "to_name",
    "to_specframe",
    "to_type",
    "to_vars",
]

# standard library
from collections.abc import Hashable
from typing import Any, TypeVar

# dependencies
import pandas as pd
import xarray as xr
from pandas.api.types import is_scalar
from typespecs import SpecFrame, from_annotated
from typespecs.typing import is_literal
from typing_extensions import get_args, get_origin

# type hints
T = TypeVar("T")


def to_attrs(specs: pd.DataFrame, /) -> dict[Any, Any]:
    """Convert given specification DataFrame to Xarray attributes."""
    attrs: dict[Any, Any] = {}

    for _, spec in specs.iterrows():
        if spec.xarray_type is None:
            type_ = lambda data: data  # type: ignore
        else:
            type_ = spec.xarray_type

        if spec.xarray_use == "attr":
            attrs[spec.xarray_name] = type_(spec.data)
        elif spec.xarray_use == "attrs":
            for name, data in spec.data.items():
                attrs[name] = type_(data)

    return attrs


def to_coords(specs: pd.DataFrame, /) -> dict[Hashable, xr.DataArray]:
    """Convert given specification DataFrame to Xarray coordinates."""
    coords: dict[Hashable, xr.DataArray] = {}

    for _, spec in specs.iterrows():
        if spec.xarray_type is None:
            type_ = xr.DataArray
        else:
            type_ = spec.xarray_type

        if spec.xarray_use == "coord":
            coords[spec.xarray_name] = type_(
                data=spec.data,
                dims=spec.xarray_dims,
                name=spec.xarray_name,
                attrs=spec.xarray_attrs,
            ).astype(  # type: ignore
                spec.xarray_dtype,
                copy=False,
            )
        elif spec.xarray_use == "coords":
            for name, data in spec.data.items():
                coords[name] = type_(
                    data=data,
                    dims=spec.xarray_dims,
                    name=name,
                    attrs=spec.xarray_attrs,
                ).astype(  # type: ignore
                    spec.xarray_dtype,
                    copy=False,
                )

    return coords


def to_data(specs: pd.DataFrame, /) -> xr.DataArray:
    """Convert given specification DataFrame to an Xarray data."""
    return next(reversed(to_vars(specs).values()))


def to_dataarray(specs: pd.DataFrame, /) -> xr.DataArray:
    """Convert given specification DataFrame to an Xarray DataArray."""
    da = to_type(specs, xr.DataArray)(to_data(specs), to_coords(specs))
    da.attrs.update(to_attrs(specs))
    da.name = to_name(specs, da.name)
    return da


def to_dataset(specs: pd.DataFrame, /) -> xr.Dataset:
    """Convert given specification DataFrame to an Xarray Dataset."""
    ds = to_type(specs, xr.Dataset)(to_vars(specs), to_coords(specs))
    ds.attrs.update(to_attrs(specs))
    return ds


def to_datatree(specs: pd.DataFrame, /) -> xr.DataTree:
    """Convert given specification DataFrame to an Xarray DataTree."""
    nodes: dict[str, xr.Dataset] = {}

    for name, group in specs.groupby("xarray_node"):
        ds = xr.Dataset(to_vars(group), to_coords(group))
        ds.attrs.update(to_attrs(group))
        nodes[name] = ds  # type: ignore

    dt = to_type(specs, xr.DataTree).from_dict(nodes)  # type: ignore
    dt.name = to_name(specs, dt.name)
    return dt


def to_dims(obj: Any, /) -> tuple[Hashable, ...] | None:
    """Convert given object to Xarray dimensions."""
    if obj is None:
        return None

    if is_literal(obj) or is_scalar(obj):
        obj = (obj,)

    if get_origin(obj) is tuple:
        obj = get_args(obj)

    return tuple(get_args(v)[0] if is_literal(v) else v for v in obj)


def to_name(specs: pd.DataFrame, default: T, /) -> T:
    """Convert given specification DataFrame to an Xarray name."""
    for _, spec in specs[::-1].iterrows():
        if spec.xarray_type is None:
            type_ = lambda data: data  # type: ignore
        else:
            type_ = spec.xarray_type

        if spec.xarray_use == "name":
            return type_(spec.data)  # type: ignore

    return default


def to_specframe(obj: Any, /) -> SpecFrame:
    """Convert given object to a specification DataFrame."""
    specs = from_annotated(
        obj,
        default={
            "xarray_attrs": None,
            "xarray_dtype": None,
            "xarray_dims": None,
            "xarray_name": None,
            "xarray_node": "/",
            "xarray_type": None,
            "xarray_use": "other",
        },
    )

    index = specs.index.to_series()
    specs["xarray_dims"] = specs["xarray_dims"].apply(to_dims)
    specs["xarray_name"] = specs["xarray_name"].fillna(index)
    return specs


def to_type(specs: pd.DataFrame, default: T, /) -> T:
    """Convert given specification DataFrame to an Xarray type."""
    for _, spec in specs[::-1].iterrows():
        if spec.xarray_type is None:
            type_ = lambda data: data  # type: ignore
        else:
            type_ = spec.xarray_type

        if spec.xarray_use == "type":
            return type_(spec.data)  # type: ignore

    return default


def to_vars(specs: pd.DataFrame, /) -> dict[Hashable, xr.DataArray]:
    """Convert given specification DataFrame to Xarray data variables."""
    vars: dict[Hashable, xr.DataArray] = {}

    for _, spec in specs.iterrows():
        if spec.xarray_type is None:
            type_ = xr.DataArray
        else:
            type_ = spec.xarray_type

        if spec.xarray_use == "data":
            vars[spec.xarray_name] = type_(
                data=spec.data,
                dims=spec.xarray_dims,
                name=spec.xarray_name,
                attrs=spec.xarray_attrs,
            ).astype(  # type: ignore
                spec.xarray_dtype,
                copy=False,
            )
        elif spec.xarray_use == "vars":
            for name, data in spec.data.items():
                vars[name] = type_(
                    data=data,
                    dims=spec.xarray_dims,
                    name=name,
                    attrs=spec.xarray_attrs,
                ).astype(  # type: ignore
                    spec.xarray_dtype,
                    copy=False,
                )

    return vars



__all__ = [
    "Dims",
    "Dtype",
    "Type",
    "attrs",
    "dims",
    "dtype",
    "name",
    "node",
    "type",
    "use",
]

# standard library
from collections.abc import Callable, Hashable, Iterable, Mapping
from typing import Annotated, Any, Literal, TypeVar, overload

# dependencies
from typespecs import ITSELF, Spec

# type hints
T = TypeVar("T")


Dims = Annotated[T, Spec(xarray_dims=ITSELF)]
"""Type hint for Xarray dimensions."""


Dtype = Annotated[T, Spec(xarray_dtype=ITSELF)]
"""Type hint for Xarray data type."""


Type = Annotated[T, Spec(xarray_type=ITSELF)]
"""Type hint for Xarray type."""


Use = Literal[
    "attr",
    "attrs",
    "coord",
    "coords",
    "data",
    "name",
    "other",
    "type",
    "vars",
]
"""Type hint for Xarray use."""


@overload
def attrs(attrs: Mapping[Any, Any] | None, /) -> Spec: ...
@overload
def attrs(**attrs: Any) -> Spec: ...
def attrs(*args: Any, **kwargs: Any) -> Spec:
    """Returns a type specification for Xarray attributes."""
    if len(args) == 0:
        return Spec(xarray_attrs=kwargs)

    if len(args) == 1 and not kwargs:
        return Spec(xarray_attrs=args[0])

    raise ValueError("Cannot create type specification.")


@overload
def dims(dims: Iterable[Hashable] | None, /) -> Spec: ...
@overload
def dims(*dims: Hashable) -> Spec: ...
def dims(*args: Any) -> Spec:
    """Returns a type specification for Xarray dimensions."""
    if len(args) == 1:
        return Spec(xarray_dims=args[0])
    else:
        return Spec(xarray_dims=args)


def dtype(dtype: Any | None, /) -> Spec:
    """Returns a type specification for Xarray data type."""
    return Spec(xarray_dtype=dtype)


def name(name: Hashable | None, /) -> Spec:
    """Returns a type specification for Xarray name."""
    return Spec(xarray_name=name)


def node(node: str, /) -> Spec:
    """Returns a type specification for Xarray node."""
    return Spec(xarray_node=node)


def type(type: Callable[..., Any] | None, /) -> Spec:
    """Returns a type specification for Xarray type."""
    return Spec(xarray_type=type)


def use(use: Use, /) -> Spec:
    """Returns a type specification for Xarray use."""
    return Spec(xarray_use=use)



__all__ = [
    # submodules
    "api",
    "convert",
    "spec",
    # aliases
    "AsDataArray",
    "AsDataset",
    "AsDataTree",
    "Dims",
    "Dtype",
    "Type",
    "asdataarray",
    "asdataset",
    "asdatatree",
    "attrs",
    "dims",
    "dtype",
    "name",
    "node",
    "type",
    "use",
]
__version__ = "0.4.0"


# dependencies
from . import api, convert, spec
from .api import *
from .spec import *



# standard library
from dataclasses import dataclass
from typing import Annotated, Any

# dependencies
import numpy as np
import xarrayspecs as xs
from numpy.typing import NDArray


@dataclass
class Temp(xs.AsDataArray):
    temp: Annotated[
        NDArray[Any],
        xs.use("data"),
        xs.dims("lon", "lat"),
        xs.dtype(np.float64),
        xs.attrs(long_name="Temperature", units="K"),
    ]
    lat: Annotated[
        NDArray[Any],
        xs.use("coord"),
        xs.dims("lat"),
        xs.dtype(np.float64),
        xs.attrs(long_name="Latitude", units="deg"),
    ]
    lon: Annotated[
        NDArray[Any],
        xs.use("coord"),
        xs.dims("lon"),
        xs.dtype(np.float64),
        xs.attrs(long_name="Longitude", units="deg"),
    ]
    date: Annotated[str, xs.use("attr")]


def test_asdataarray() -> None:
    Temp.new(
        np.random.uniform(273, 293, size=(2, 2)),
        np.array([0, 1]),
        np.array([2, 3]),
        "2026-03-01",
    )



# standard library
from dataclasses import dataclass
from typing import Annotated, Any

# dependencies
import numpy as np
import xarrayspecs as xs
from numpy.typing import NDArray

# type hints
Temp = Annotated[
    NDArray[Any],
    xs.use("data"),
    xs.name("temp"),
    xs.dims("lon", "lat"),
    xs.dtype(np.float64),
    xs.attrs(long_name="Temperature", units="K"),
]
Humid = Annotated[
    NDArray[Any],
    xs.use("data"),
    xs.name("humid"),
    xs.dims("lon", "lat"),
    xs.dtype(np.float64),
    xs.attrs(long_name="Humidity", units="%"),
]
Lat = Annotated[
    NDArray[Any],
    xs.use("coord"),
    xs.name("lat"),
    xs.dims("lat"),
    xs.dtype(np.float64),
    xs.attrs(long_name="Latitude", units="deg"),
]
Lon = Annotated[
    NDArray[Any],
    xs.use("coord"),
    xs.name("lon"),
    xs.dims("lon"),
    xs.dtype(np.float64),
    xs.attrs(long_name="Longitude", units="deg"),
]
Date = Annotated[str, xs.use("attr"), xs.name("date")]


@dataclass
class Weathers(xs.AsDataTree):
    temp_0: Annotated[Temp, xs.node("/0")]
    temp_1: Annotated[Temp, xs.node("/1")]
    humid_0: Annotated[Humid, xs.node("/0")]
    humid_1: Annotated[Humid, xs.node("/1")]
    lat_0: Annotated[Lat, xs.node("/0")]
    lat_1: Annotated[Lat, xs.node("/1")]
    lon_0: Annotated[Lon, xs.node("/0")]
    lon_1: Annotated[Lon, xs.node("/1")]
    date_0: Annotated[Date, xs.node("/0")]
    date_1: Annotated[Date, xs.node("/1")]


def test_asdatatree() -> None:
    Weathers.new(
        np.random.uniform(273, 293, size=(2, 2)),
        np.random.uniform(273, 293, size=(2, 2)),
        np.random.uniform(0, 100, size=(2, 2)),
        np.random.uniform(0, 100, size=(2, 2)),
        np.arange(2),
        np.arange(2) + 1,
        np.arange(2),
        np.arange(2) + 1,
        "2026-03-01",
        "2026-03-01",
    )
